"""
Unit and integration tests for tests `cappr.openai.classify`.
TODO: this should instead be structured like test_huggingface_classify.py!
"""
from __future__ import annotations
import os
import re
import sys

import numpy as np
import pandas as pd
import pytest
import tiktoken

from cappr import Example as Ex
from cappr.openai import classify

# sys hack to import from parent. If someone has a cleaner solution, lmk
sys.path.insert(1, os.path.join(sys.path[0], ".."))
import _test


@pytest.fixture(autouse=True)
def patch_openai_method_retry(monkeypatch):
    # During testing, there's never going to be a case where we want to actually hit
    # an OpenAI endpoint!
    def _log_probs(texts: list[str]) -> list[list[float]]:
        """
        Returns a list `log_probs` where `log_probs[i]` is `list(range(size))` where
        `size` is the number of tokens in `texts[i]`.
        """
        tokenizer = tiktoken.get_encoding("gpt2")
        return [list(range(len(tokens))) for tokens in tokenizer.encode_batch(texts)]

    # Note that the the text completion endpoint uses a kwarg named prompt, but that's
    # actually set to prompt + completion in the CAPPr scheme. We input that to get
    # log-probs of completion tokens given prompt
    def mocked(openai_method, prompt: list[str], **kwargs):
        # Technically, we should return a openai.openai_object.OpenAIObject
        # For now, just gonna return the minimum dict required
        token_logprobs_batch = _log_probs(prompt)
        return {
            "choices": [
                {"logprobs": {"token_logprobs": list(token_logprobs)}}
                for token_logprobs in token_logprobs_batch
            ]
        }

    monkeypatch.setattr("cappr.openai.api.openai_method_retry", mocked)


@pytest.fixture(autouse=True)
def patch_tokenizer(monkeypatch):
    monkeypatch.setattr(
        "tiktoken.encoding_for_model", lambda _: tiktoken.get_encoding("gpt2")
    )


@pytest.fixture(scope="module")
def model():
    """
    This name is intentionally *not* an OpenAI API model. That's to prevent un-mocked
    API calls from going through.
    """
    return "🦖 ☄️ 💥"


@pytest.fixture(scope="module")
def prompts():
    return [
        "Fill in the blank. Have an __ day!",
        "i before e except after",
        "Popular steak sauce:",
        "The English alphabet: a, b,",
    ]


@pytest.fixture(scope="module")
def completions():
    return ["A1", "c"]


@pytest.fixture(scope="module")
def examples():
    # Let's make these ragged (different # completions per prompt), since
    # that's a use case for an Example
    return [
        Ex(prompt="lotta", completions=("media", "food"), prior=(1 / 2, 1 / 2)),
        Ex(
            prompt=("🎶The best time to wear a striped sweater is all the"),
            completions=("time🎶",),
            normalize=False,
        ),
        Ex(
            prompt="machine",
            completions=("-washed", " learnt", " 🤖"),
            prior=(1 / 6, 2 / 3, 1 / 6),
            end_of_prompt="",
        ),
    ]


def test_token_logprobs(model):
    texts = ["a b c", "d e"]
    log_probs = classify.token_logprobs(texts, model)
    assert log_probs == [[0, 1, 2], [0, 1]]  # cuz the API is mocked w/ range(len())


def test__slice_completions(completions, model):
    log_probs = [[0, 1, 2], [0, 1]]
    log_probs_completions = classify._slice_completions(
        completions, end_of_prompt="", log_probs=log_probs, model=model
    )
    assert log_probs_completions == [[1, 2], [1]]


def test_log_probs_conditional(prompts, completions, model):
    expected = [[[10, 11], [10]], [[5, 6], [5]], [[5, 6], [5]], [[8, 9], [8]]]

    # Test plain input
    _log_probs_conditional = classify.log_probs_conditional(prompts, completions, model)
    assert _log_probs_conditional == expected

    # Test that you can input a pandas Series w/ an arbitrary index
    _prompts_series = pd.Series(
        prompts, index=np.random.choice(len(prompts), size=len(prompts))
    )
    _completions_series = pd.Series(
        completions, index=np.random.choice(len(completions), size=len(completions))
    )
    _log_probs_conditional = classify.log_probs_conditional(
        _prompts_series, _completions_series, model
    )
    assert _log_probs_conditional == expected

    # Test bad prompts - empty
    with pytest.raises(ValueError, match="prompts must be non-empty."):
        classify.log_probs_conditional([], completions, model)

    # Test bad prompts - non-ordered
    with pytest.raises(TypeError, match="prompts must be an ordered collection."):
        classify.log_probs_conditional(set(prompts), completions, model)

    # Test bad completions - empty
    with pytest.raises(ValueError, match="completions must be non-empty."):
        classify.log_probs_conditional(prompts, [], model)

    # Test bad completions - non-ordered
    with pytest.raises(TypeError, match="completions must be an ordered collection."):
        classify.log_probs_conditional(prompts, set(completions), model)

    # Test bad completions - string
    with pytest.raises(TypeError, match="completions cannot be a string."):
        classify.log_probs_conditional(prompts, completions[0], model)


def test_log_probs_conditional_examples(examples, model):
    log_probs_conditional = classify.log_probs_conditional_examples(examples, model)
    expected = [[[2], [2]], [[14, 15, 16, 17]], [[1, 2], [1], [1, 2, 3]]]
    assert log_probs_conditional == expected

    # Test bad examples - non-ordered
    with pytest.raises(TypeError, match="examples must be an ordered collection."):
        classify.log_probs_conditional_examples(set(examples), model)

    # Test bad examples - empty
    with pytest.raises(ValueError, match="examples must be non-empty."):
        classify.log_probs_conditional_examples([], model)


def test_predict_proba(prompts, completions, model):
    _test.predict_proba(classify.predict_proba, prompts, completions, model)

    # Test prior
    prior = [1 / len(completions)] * len(completions)
    _test.predict_proba(
        classify.predict_proba, prompts, completions, model, prior=prior
    )

    # Test discount_completions > 0.0
    _test.predict_proba(
        classify.predict_proba, prompts, completions, model, discount_completions=1.0
    )

    # Test discount_completions > 0.0 with prior
    _test.predict_proba(
        classify.predict_proba,
        prompts,
        completions,
        model,
        prior=prior,
        discount_completions=1.0,
    )

    # Test bad prior input - wrong size
    prior_bad = prior + [0]
    expected_error_msg = re.escape(
        f"Expected prior to have length {len(completions)} (the number of "
        f"completions), got {len(prior_bad)}."
    )
    with pytest.raises(ValueError, match=expected_error_msg):
        classify.predict_proba(prompts, completions, model, prior=prior_bad)

    # Test bad prior input - not a probability distr
    prior_bad = prior[:-1] + [0]
    expected_error_msg = "prior must sum to 1."
    with pytest.raises(ValueError, match=expected_error_msg):
        classify.predict_proba(prompts, completions, model, prior=prior_bad)

    # Test bad normalize input
    expected_error_msg = "Setting normalize=True when there's only 1 completion"
    with pytest.raises(ValueError, match=expected_error_msg):
        classify.predict_proba(prompts, completions[:1], model, normalize=True)


def test_predict_proba_examples(examples: list[Ex], model):
    _test.predict_proba_examples(classify.predict_proba_examples, examples, model)


def test_predict(prompts, completions, model):
    _test.predict(classify.predict, prompts, completions, model)

    # Test that you can input a pandas Series w/ an arbitrary index
    _prompts_series = pd.Series(
        prompts, index=np.random.choice(len(prompts), size=len(prompts))
    )
    _completions_series = pd.Series(
        completions, index=np.random.choice(len(completions), size=len(completions))
    )
    _test.predict(classify.predict, _prompts_series, _completions_series, model)

    # test discount_completions > 0.0
    _test.predict(
        classify.predict, prompts, completions, model, discount_completions=1.0
    )


def test_predict_examples(examples: list[Ex], model):
    _test.predict_examples(classify.predict_examples, examples, model)



"""
Batch lists into sublists of constant or variable sizes, and batchify functions.
"""
from __future__ import annotations
from functools import wraps
import inspect
from typing import Any, Optional, Sequence

import numpy as np
from tqdm.auto import tqdm


MIN_TOTAL_FOR_SHOWING_PROGRESS_BAR = 5


def constant(lst: list, size: int):
    """
    Generates sublists in the order of `lst` which partition `lst`. All sublists (except
    potentially the last) have length `size`.
    """
    if size <= 0:
        raise ValueError("size must be positive.")
    lst = list(lst)  # 0-index whatever was passed, or fully evaluate generator
    for idx in range(0, len(lst), size):
        yield lst[idx : (idx + size)]


def variable(lst: list, sizes: Sequence[int]):
    """
    Generates sublists in the order of `lst` which partition `lst`. The `i`'th generated
    sublist has length `sizes[i]`.
    """
    sizes: np.ndarray = np.array(sizes)
    if np.any(sizes <= 0):
        raise ValueError("sizes must all be positive.")
    if len(sizes.shape) != 1:
        raise ValueError("sizes must be 1-D.")
    cumulative_sizes = np.cumsum(sizes)
    if cumulative_sizes[-1] != len(lst):
        raise ValueError("sizes must sum to len(lst).")
    # We want start and stop idxs. The first start idx must ofc be 0.
    cumulative_sizes = np.insert(cumulative_sizes, 0, 0)
    lst = list(lst)  # 0-index and/or fully evaluate generator
    for start, stop in zip(cumulative_sizes[:-1], cumulative_sizes[1:]):
        yield lst[start:stop]


def _kwarg_name_to_value(func) -> dict[str, Any]:
    """
    Returns a dictionary mapping keyword arguments to their default values `func`'s
    signature.
    """
    # ty https://stackoverflow.com/a/12627202/18758987
    signature = inspect.signature(func)
    return {
        name: value.default
        for name, value in signature.parameters.items()
        if value.default is not inspect.Parameter.empty
    }


def batchify(
    batchable_arg: str,
    batch_size: int = 32,
    progress_bar_desc: str = "",
    show_progress_bar: Optional[bool] = None,
):
    """
    Returns a decorator which runs the decorated function in batches along its
    `batchable_arg`, returning a list of the function's outputs for each batch.

    If the function includes a `'batch_size'` keyword argument, then its value is used
    as the batch size instead of the decorator's default `batch_size`.
    """

    def decorator(func):
        _arg_names = inspect.getfullargspec(func).args
        batchable_arg_idx = _arg_names.index(batchable_arg)
        _kwargs_signature = _kwarg_name_to_value(func)
        batch_size_default: int = _kwargs_signature.get("batch_size", batch_size)
        show_progress_bar_default: Optional[bool] = _kwargs_signature.get(
            "show_progress_bar", show_progress_bar
        )

        @wraps(func)
        def wrapper(*args, **kwargs):
            # set up batching
            batchable: Sequence = args[batchable_arg_idx]
            batch_size: int = kwargs.get("batch_size", batch_size_default)
            args = list(args)  # needs to be mutable to modify the batch argument value
            # set up progress bar
            show_progress_bar: Optional[bool] = kwargs.get(
                "show_progress_bar", show_progress_bar_default
            )
            total = len(batchable)
            disable = not (
                show_progress_bar or total >= MIN_TOTAL_FOR_SHOWING_PROGRESS_BAR
            )
            # run func along batches of batchable
            outputs = []
            with tqdm(
                total=total, desc=progress_bar_desc, disable=disable
            ) as progress_bar:
                for batch_ in constant(batchable, batch_size):
                    args[batchable_arg_idx] = batch_
                    outputs.append(func(*args, **kwargs))
                    progress_bar.update(len(batch_))
            return outputs

        return wrapper

    return decorator


def flatten(batchified_func):
    """
    Decorates a `cappr.utils._batch.batchify`d function. Flattens the output.
    """

    @wraps(batchified_func)
    def wrapper(*args, **kwargs):
        nested_outputs = batchified_func(*args, **kwargs)
        return [output for inner_outputs in nested_outputs for output in inner_outputs]

    return wrapper



"""
Unit tests `cappr.openai.api`. Currently pretty barebones.
"""
from __future__ import annotations

import openai
import pytest

from cappr.openai import api


def test_openai_method_retry():
    def openai_method(**kwargs):
        return kwargs["waddup"]

    assert api.openai_method_retry(openai_method, waddup="nada") == "nada"

    def openai_method():
        raise openai.error.ServiceUnavailableError

    with pytest.raises(openai.error.ServiceUnavailableError):
        api.openai_method_retry(openai_method, sleep_sec=0)


def test__openai_api_call_is_ok(monkeypatch):
    # Inputs
    texts = ["cherry", "coke"]  # tokenized: [[331, 5515], [1030, 441]]
    model = "gpt-3.5-turbo-instruct"  # hard-coded so that tokenization is always same
    max_tokens = 3  # number of completion tokens per prompt
    cost_per_1k_tokens_prompt = 1  # make it somewhat big to avoid numerical issues
    cost_per_1k_tokens_completion = 2

    # Expected outputs
    num_tokens_prompts_expected = 4
    num_tokens_completions_expected = len(texts) * max_tokens
    cost_expected = round(
        (
            (num_tokens_prompts_expected * cost_per_1k_tokens_prompt)
            + (num_tokens_completions_expected * cost_per_1k_tokens_completion)
        )
        / 1_000,
        2,
    )

    # Mimic the user seeing the prompt and entering y
    monkeypatch.setattr("builtins.input", lambda _: "y")
    (
        num_tokens_prompts_observed,
        num_tokens_completions_observed,
        cost_observed,
    ) = api._openai_api_call_is_ok(
        texts,
        model,
        max_tokens=max_tokens,
        cost_per_1k_tokens_prompt=cost_per_1k_tokens_prompt,
        cost_per_1k_tokens_completion=cost_per_1k_tokens_completion,
    )
    assert num_tokens_completions_observed == num_tokens_completions_expected
    assert num_tokens_prompts_observed == num_tokens_prompts_expected
    assert cost_observed == cost_expected

    # Mimic the user seeing the prompt and entering n
    monkeypatch.setattr("builtins.input", lambda _: "n")
    with pytest.raises(api._UserCanceled):
        api._openai_api_call_is_ok(texts, model)


def test_gpt_chat_complete(monkeypatch):
    completion_expected = "heyteam howsitgoin"

    def mocked(openai_method, messages: list[dict[str, str]], **kwargs):
        # Technically, we should return a openai.openai_object.OpenAIObject
        # For now, just gonna return the minimum dict required
        return {"choices": [{"text": completion_expected}]}

    monkeypatch.setattr("cappr.openai.api.openai_method_retry", mocked)

    prompts = ["hey there", "hi", "hmm"]
    choices = api.gpt_chat_complete(prompts, model="o_o hi")
    completions = [choice["text"] for choice in choices]
    assert len(prompts) == len(completions)
    for completion_observed in completions:
        assert completion_observed == completion_expected



"""
Helper functions to test that `predict_proba` and `predict` function outputs have the
correct form.
"""
from __future__ import annotations
from typing import Any, Callable, Sequence, Union

import numpy as np
import numpy.typing as npt

from cappr import Example


def predict_proba(
    predict_proba_func: Callable[
        [Union[str, Sequence[str]], Sequence[str], Any], npt.NDArray[np.floating]
    ],
    prompts: Union[str, Sequence[str]],
    completions: Sequence[str],
    *args,
    **kwargs,
):
    """
    Tests that `predict_proba_func(prompts, completions, *args, **kwargs)` returns a
    numpy array with the correct shape, and that each row is a probability distribution
    if `normalize=True`.
    """
    pred_probs = predict_proba_func(prompts, completions, *args, **kwargs)
    assert isinstance(pred_probs, np.ndarray)
    if isinstance(prompts, str):
        assert pred_probs.shape == (len(completions),)
        completions_axis = 0
    else:
        assert pred_probs.shape == (len(prompts), len(completions))
        completions_axis = 1
    assert np.all(pred_probs >= 0)
    assert np.all(pred_probs <= 1)

    if pred_probs.shape[-1] == 1:
        if isinstance(prompts, str):
            p = pred_probs
        else:
            p = pred_probs[:, 0]
        # We don't normalize if there's one completion. It's almost definitely the case
        # that the predicted probability is < 1 by a decent amount.
        assert not any(np.isclose(p, 1))
    else:
        prob_sums = pred_probs.sum(axis=completions_axis)
        are_close_to_1 = np.isclose(prob_sums, 1)
        if kwargs.get("normalize", True):
            assert np.all(are_close_to_1)
        else:
            assert not np.any(are_close_to_1)


def predict_proba_examples(
    predict_proba_examples_func: Callable[
        [Union[Example, Sequence[Example]], Any],
        Union[npt.NDArray[np.floating], list[npt.NDArray[np.floating]]],
    ],
    examples: Union[Example, Sequence[Example]],
    *args,
    **kwargs,
):
    """
    Tests that `predict_proba_examples_func(examples, *args, **kwargs)` returns a
    list with the correct shape, and that each element is a probability distribution.
    """

    def test_single_example(
        pred_prob_example: npt.NDArray[np.floating], example: Example
    ):
        assert pred_prob_example.shape == (len(example.completions),)
        assert np.all(pred_prob_example >= 0)
        if len(example.completions) > 1:
            # Testing artifact: for mocked log-prob API endpoints, we currently return
            # integers, not log-probs. This fact combined with the fact that we obv
            # don't normalize when there's 1 completion causes the test "probability"
            # output to be > 1. So let's only run this test for examples w/ more than 1
            # completion
            assert np.all(pred_prob_example <= 1)
        if len(example.completions) == 1:
            # We don't normalize if there's one completion. It's almost definitely the
            # case that the predicted probability is < 1 by a decent amount.
            assert not any(np.isclose(pred_prob_example, 1))
        else:
            prob_sum = pred_prob_example.sum()
            is_close_to_1 = np.isclose(prob_sum, 1)
            if example.normalize:
                assert is_close_to_1
            else:
                assert not is_close_to_1

    pred_probs = predict_proba_examples_func(examples, *args, **kwargs)
    if isinstance(examples, Example):
        test_single_example(pred_probs, examples)
    else:
        assert len(pred_probs) == len(examples)
        for pred_prob_example, example in zip(pred_probs, examples):
            test_single_example(pred_prob_example, example)


def predict(
    predict_func: Callable[
        [Union[str, Sequence[str]], Sequence[str], Any], Union[str, list[str]]
    ],
    prompts: Union[str, Sequence[str]],
    completions: Sequence[str],
    *args,
    **kwargs,
):
    """
    Tests that `predict_func(prompts, completions, *args, **kwargs)` returns an array
    with length `len(prompts)`, and that each element is in `completions`.
    """
    preds = predict_func(prompts, completions, *args, **kwargs)
    # If completions is a pandas Series, then __contains__ checks the Series index, not
    # the values! Just convert to tuple.
    completions = tuple(completions)
    if isinstance(prompts, str):
        assert isinstance(preds, str)
        assert preds in completions
    else:
        assert len(preds) == len(prompts)
        assert all([pred in completions for pred in preds])


def predict_examples(
    predict_examples_func: Callable[
        [Union[Example, Sequence[Example]], Any], Union[str, list[str]]
    ],
    examples: Union[Example, Sequence[Example]],
    *args,
    **kwargs,
):
    """
    Tests that `predict_examples_func(examples, *args, **kwargs)` returns an array
    with length `len(examples)`, and that each element is one of the example's
    completions.
    """
    preds = predict_examples_func(examples, *args, **kwargs)
    if isinstance(examples, Example):
        assert isinstance(preds, str)
        assert preds in examples.completions
    else:
        assert len(preds) == len(examples)
        for pred, example in zip(preds, examples):
            assert pred in example.completions



"""
Completion After Prompt Probability
"""
__version__ = "0.4.6"

from . import utils
from ._example import Example
from . import openai

try:
    from . import huggingface  # optional
except ModuleNotFoundError:
    pass



"""
YouTils
"""
from __future__ import annotations
from contextlib import contextmanager
from typing import Sequence, TypeVar

import torch
import torch.nn.functional as F
from transformers import (
    BatchEncoding,
    LlamaTokenizer,
    LlamaTokenizerFast,
    PreTrainedModel,
    PreTrainedTokenizerBase,
)


# AutoModelForCausalLM is not actually a superclass for pretrained causal/autoregressive
# LMs. Its from_pretrained method is a loading utility which always returns a
# PreTrainedModel. Let's create a type for them to make the documentation clearer.
PreTrainedModelForCausalLM = TypeVar(
    "PreTrainedModelForCausalLM", bound=PreTrainedModel
)
"A model loaded via `transformers.AutoModelForCausalLM.from_pretrained`"


@contextmanager
def set_up_model_and_tokenizer(
    model_and_tokenizer: tuple[PreTrainedModelForCausalLM, PreTrainedTokenizerBase]
):
    """
    In this context, internal attributes of the model and tokenizer are set to enable
    correct, batched inference. Namely:
      - the model is set in eval mode
      - the tokenizer pads on the right
      - the tokenizer does not add an EOS token.
    """
    model, tokenizer = model_and_tokenizer

    # Grab attributes - model
    model_is_train = model.training
    # Grab attributes - tokenizer
    tokenizer_pad_token_id = tokenizer.pad_token_id
    tokenizer_padding_side = tokenizer.padding_side
    if hasattr(tokenizer, "add_eos_token"):
        tokenizer_add_eos_token = tokenizer.add_eos_token

    # Set attributes - model
    # 1. Just ensure that the model is in eval mode. CAPPr only makes sense as an
    #    inference computation.
    model.eval()

    # Set attributes - tokenizer
    # Note: PreTrainedTokenizerBase is smart about setting auxiliary attributes, e.g.,
    # it updates tokenizer.special_tokens_map after setting tokenizer.pad_token_id.
    # 1. Set the pad token (if it's not set) so that batch inference is possible. These
    #    get masked out. Keep in mind that you need to be careful about setting position
    #    IDs correctly.
    if tokenizer_pad_token_id is None:
        tokenizer.pad_token_id = tokenizer.eos_token_id
    # 2. Set the padding side to right. Left-padding would alter position IDs for
    #    non-pad tokens, which makes things a bit more confusing.
    tokenizer.padding_side = "right"
    # 3. Don't add an end-of-sentence token. We'll never need it for the CAPPr scheme.
    #    Keeping it would throw off the classify module (which caches).
    if hasattr(tokenizer, "add_eos_token"):
        tokenizer.add_eos_token = False

    yield

    # Reset attributes - model
    model.train(model_is_train)
    # Reset attributes - tokenizer
    if tokenizer_pad_token_id is None:
        tokenizer.pad_token_id = tokenizer_pad_token_id
    tokenizer.padding_side = tokenizer_padding_side
    if hasattr(tokenizer, "add_eos_token"):
        tokenizer.add_eos_token = tokenizer_add_eos_token


@contextmanager
def disable_add_bos_token(tokenizer: PreTrainedTokenizerBase):
    if hasattr(tokenizer, "add_bos_token"):
        add_bos_token: bool = tokenizer.add_bos_token
        tokenizer.add_bos_token = False

    yield

    if hasattr(tokenizer, "add_bos_token"):
        tokenizer.add_bos_token = add_bos_token


def logits_texts(
    texts: Sequence[str],
    model: PreTrainedModelForCausalLM,
    tokenizer: PreTrainedTokenizerBase,
) -> tuple[torch.Tensor, BatchEncoding]:
    # TODO: auto-batch? consider adding a batch_size kwarg, and decorating the func like
    # token_logprobs
    encodings = tokenizer(texts, return_tensors="pt", padding=True).to(model.device)
    with torch.no_grad():
        out = model(**encodings)
    if getattr(tokenizer, "add_bos_token", False):
        # Drop the first bos token after we're done encoding so that the shape is
        # consistent w/ other tokenizers
        logits = out.logits[:, 1:, :]
        encodings = BatchEncoding(
            {key: value[:, 1:] for key, value in encodings.items()}
        )
    else:
        logits = out.logits
    return logits, encodings


def logits_to_log_probs(
    logits: torch.Tensor,
    input_ids: torch.Tensor,
    input_ids_start_idx: int = None,
    logits_end_idx: int = None,
):
    """
    TODO: docstring
    """
    # logits.shape is (# texts, max # tokens in texts, vocab size)
    log_probs = F.log_softmax(logits, dim=2)

    # Only keep the log-prob from the vocab dimension whose index is is the
    # next token's input ID.
    # input_ids.shape is (# texts, max # tokens in texts)
    return (
        log_probs[:, :logits_end_idx, :]
        .take_along_dim(input_ids[:, input_ids_start_idx:, None], dim=2)
        .squeeze(-1)
    )


def does_tokenizer_prepend_space_to_first_token(
    tokenizer: PreTrainedTokenizerBase,
) -> bool:
    return not isinstance(tokenizer, (LlamaTokenizer, LlamaTokenizerFast))



"""
Shared input checks. These check conditions that would cause silent and annoying
failures. For example, inputting an unordered iterable will cause outputs like predicted
probabilities to be meaningless.
"""
from __future__ import annotations
from typing import Optional, Sequence

import numpy as np


def _is_reversible(object) -> bool:
    # First, cheaply check if it implements __reversed__
    if hasattr(object, "__reversed__"):
        return True
    # Some objects don't have the attribute, but still can be reversible, like a tuple.
    # reversed(object) is often a generator, so checking this is often cheap.
    try:
        reversed(object)
    except TypeError:
        return False
    else:
        return True


def ordered(object: Sequence, variable_name: str):
    # This check isn't perfect, but it works well enough.
    if not _is_reversible(object):
        raise TypeError(
            f"{variable_name} must be an ordered collection. Consider converting it to "
            "a list or tuple."
        )


def nonempty(object: Sequence, variable_name: str):
    if len(object) == 0:
        raise ValueError(f"{variable_name} must be non-empty.")


def nonempty_and_ordered(object: Sequence, variable_name: str):
    nonempty(object, variable_name)
    ordered(object, variable_name)


def completions(completions: Sequence[str]):
    """
    Raise an error if `completions` is not a nonempty, ordered, non-string.
    """
    nonempty_and_ordered(completions, variable_name="completions")
    if isinstance(completions, str):
        raise TypeError(
            "completions cannot be a string. It must be a sequence of strings. If you "
            "intend on inputting a single completion to estimate its probability, wrap "
            "it in a list or tuple and set normalize=False."
        )


def prior(prior: Optional[Sequence[float]], expected_length: int):
    """
    Raises an error if `prior is not None` and isn't a probability distribution over
    `expected_length` categories.
    """
    if prior is None:  # it's a uniform prior, no need to check anything
        return
    # Going to be stricter on the type here. We do the prior * likelihood computation
    # after making expensive model calls. Don't want that multiplication to fail b/c
    # that'd be a complete waste of model compute.
    if not isinstance(prior, (Sequence, np.ndarray)):
        raise TypeError("prior must be None, a Sequence, or numpy array.")
    if len(np.shape(prior)) != 1:
        raise ValueError("prior must be 1-D.")
    prior_arr = np.array(prior, dtype=float)  # try casting to float
    if not np.all(prior_arr >= 0):
        raise ValueError("prior must contain probabilities between 0 and 1.")
    if not np.all(prior_arr <= 1):
        raise ValueError("prior must contain probabilities between 0 and 1.")
    if not np.isclose(prior_arr.sum(), 1):
        raise ValueError("prior must sum to 1.")
    if prior is not None and len(prior) != expected_length:
        raise ValueError(
            f"Expected prior to have length {expected_length} (the number of "
            f"completions), got {len(prior)}."
        )


def normalize(completions: Sequence[str], normalize: bool):
    """
    Raises a `ValueError` if `len(completions) == 1 and normalize`.
    """
    if len(completions) == 1 and normalize:
        raise ValueError(
            "Setting normalize=True when there's only 1 completion causes the "
            "probability to trivially be 1. Did you mean to set normalize=False, or "
            "did you mean to include more completions?"
        )
